/*
   Copyright (c) 1995-1996 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute this software for
   any purpose and without fee is hereby granted, provided that this
   copyright and permission notice appear on all copies of the
   software and supporting documentation, the name of Cisco Systems,
   Inc. not be used in advertising or publicity pertaining to
   distribution of the program without specific prior permission, and
   notice be given in supporting documentation that modification,
   copying and distribution is by permission of Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <stdlib.h>
#include <strings.h>
#include "tac_plus.h"
#include "expire.h"

#ifdef HAVE_SHADOW_H
#include <shadow.h>
#endif

#ifdef PWDDB
#include "pwdt.h"
#endif

#ifdef TAC_SMB_AUTH
#include "../lib/libasmb/include/pwcheck_smb.h"
extern void cfg_get_smb_data(char *, int, char **, char **, char **);
#endif
#ifdef TAC_LDAP_AUTH
extern int pwcheck_ldap(char *, char *, char *, char *, char *, char *, char *, char *, char *);
extern void cfg_get_ldap_data(char *, int, char **, char **, char **, char **
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    , char **, char **
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    , char **
#endif
);
#endif
#ifdef TAC_DCE_AUTH
extern int pwcheck_dce(char *userid, char *passwd)
#endif

/* Generic password verification routines for des, file and cleartext
   passwords */

static int passwd_file_verify();

/* Adjust data->status depending on whether a user has expired or not */

void
set_expiration_status(exp_date, data)
char *exp_date;
struct authen_data *data;
{
    int expired;

    /* if the status is anything except pass, there's no point proceeding */
    if (data->status != TAC_PLUS_AUTHEN_STATUS_PASS) {
	return;
    }

    /* Check the expiration date, if any. If NULL, this check will return
     * PW_OK */
    expired = check_expiration(exp_date);

    switch (expired) {
    case PW_OK:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password has not expired %s",
		   exp_date ? exp_date : "<no expiry date set>");

	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	return;

    case PW_EXPIRING:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password will expire soon %s",
		   exp_date ? exp_date : "<no expiry date set>");
	if (data->server_msg)
	    free(data->server_msg);
	data->server_msg = tac_strdup("Password will expire soon");
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	return;

    case PW_EXPIRED:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password has expired %s",
		   exp_date ? exp_date : "<no expiry date set>");
	if (data->server_msg)
	    free(data->server_msg);
	data->server_msg = tac_strdup("Password has expired");
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return;

    default:
	report(LOG_ERR, "%s: Bogus return value %d from check_expiration",
	       session.peer, expired);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }
}

/* Verify that this user/password is valid.  Works only for cleartext,
   file and des passwords.

   Return 1 if password is valid */

int
verify(name, passwd, data, recurse)
char *name, *passwd;
struct authen_data *data;
int recurse;
{
    char *exp_date;
    char *cfg_passwd;
    char *p;

#ifdef PERUSER_ENABLE
    if(data->service == TAC_PLUS_AUTHEN_SVC_ENUSER)
      cfg_passwd = (char*)cfg_get_enable (name, recurse);
    else {
#endif

    if (data->type == TAC_PLUS_AUTHEN_TYPE_PAP) {
	cfg_passwd = cfg_get_pap_secret(name, recurse);
    } else {
	cfg_passwd = cfg_get_login_secret(name, recurse);
    }

    /* If there is no login or pap password for this user, see if there is
       a global password for her that we can use */

    if (!cfg_passwd) {
	cfg_passwd = cfg_get_global_secret(name, recurse);
    }

    /* If we still have no password for this user (or no user for that
       matter) but the default authentication = file <file> statement
       has been issued, attempt to use this password file */

    if (!cfg_passwd) {
	char *file = cfg_get_authen_default();
	if (file) {
/*	    return passwd_file_verify(name, passwd, data, file, TAC_PLUS_NORECURSE);
 */
	    return passwd_file_verify(name, passwd, data, file, recurse);
	}

	/* otherwise, we fail */
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }

#ifdef PERUSER_ENABLE
    }
#endif

    /* We have a configured password. Deal with it depending on its
       type */

#if defined(NONE_LOGIN) || defined(PERUSER_ENABLE)
    if(STREQ("none", cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	exp_date = cfg_get_expires(name, recurse);
	set_expiration_status(exp_date, data);
	if(!sys_adm_verify(name, 0, data, recurse))
	  return 0;
	return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
    }
    /* theoretically impossible cases, but... */
    if(STREQ("permit", cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	return 1;
    }
    if(STREQ("deny", cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return 0;
    }
#endif

    p = tac_find_substring("cleartext ", cfg_passwd);
    if (p) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify daemon %s == NAS %s", p, passwd);

	if (strcmp(passwd, p)) {
	    if (debug & DEBUG_PASSWD_FLAG)
		report(LOG_DEBUG, "Password is incorrect");
	    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	    return(0);
	} else {
	    data->status = TAC_PLUS_AUTHEN_STATUS_PASS;

	    if (debug & DEBUG_PASSWD_FLAG)
		report(LOG_DEBUG, "Password is correct");
	}

	exp_date = cfg_get_expires(name, recurse);
	set_expiration_status(exp_date, data);
	if(!sys_adm_verify(name, 0, data, recurse))
	  return 0;
	return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
    }

    p = tac_find_substring("des ", cfg_passwd);
    if (p) {
	/* try to verify this des password */
	if (!des_verify(passwd, p)) {
	    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	    return (0);
	} else {
	    data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	}

	exp_date = cfg_get_expires(name, recurse);
	set_expiration_status(exp_date, data);
	if(!sys_adm_verify(name, 0, data, recurse))
	  return 0;
	return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
    }

    p = tac_find_substring("file ", cfg_passwd);
    if (p) {
	return passwd_file_verify(name, passwd, data, p, recurse);
    }

#ifdef TAC_SMB_AUTH
    p = tac_find_substring("smb", cfg_passwd);
    if (p) {
	char *smbprimary, *smbbackup, *smbdomain;
	int res;

	cfg_get_smb_data(name, recurse, &smbprimary, &smbbackup, &smbdomain);
	res = pwcheck_smb(name, passwd, smbprimary, smbbackup, smbdomain);
	data->status = (res == NTV_NO_ERROR) ? TAC_PLUS_AUTHEN_STATUS_PASS : TAC_PLUS_AUTHEN_STATUS_FAIL;
	switch (res) {
	    case NTV_NO_ERROR:
		report(LOG_INFO, "smb auth ok for %s", name);
		break;
	    case NTV_SERVER_ERROR:
		report(LOG_ERR, "smb auth: server error for %s", name);
		break;
	    case NTV_PROTOCOL_ERROR:
		report(LOG_ERR, "smb auth: protocol error for %s", name);
		break;
	    case NTV_LOGON_ERROR:
		report(LOG_ERR, "smb auth: logon error for %s", name);
		break;
	    default:
		report(LOG_ERR, "smb auth: unknown error %i for %s", res, name);
		break;
	}
	return (res == NTV_NO_ERROR);
    }
#endif

#ifdef TAC_LDAP_AUTH
    p = tac_find_substring("ldap", cfg_passwd);
    if (p) {
	char *ldapserver, *ldapbasedn, *ldapbinddn, *ldapbindpw;
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
	char *ldapsaslmech, *ldapsaslcred;
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
	char *ldapsslpath;
#endif
	int res;

	cfg_get_ldap_data(name, recurse, &ldapserver, &ldapbinddn, &ldapbindpw, &ldapbasedn
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
	    , &ldapsaslmech, &ldapsaslcred
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
	    , &ldapsslpath
#endif
);
	res = pwcheck_ldap(name, passwd, ldapbasedn, ldapserver, ldapbinddn, ldapbindpw
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
	    , ldapsaslmech, ldapsaslcred
#else
	    , NULL, NULL
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
	    , ldapsslpath
#else
	    , NULL
#endif
);
	data->status = (res == 1) ? TAC_PLUS_AUTHEN_STATUS_PASS : TAC_PLUS_AUTHEN_STATUS_FAIL;
	if (!res)
	    report(LOG_ERR, "ldap auth error for %s", name);
	else
	    report(LOG_INFO, "ldap auth ok for %s", name);
	return (res == 1);
    }
#endif

#ifdef TAC_DCE_AUTH
    p = tac_find_substring("dce", cfg_passwd);
    if (p) {
	int res;

	res = pwcheck_dce(name, passwd);
	data->status = (res == 1) ? TAC_PLUS_AUTHEN_STATUS_PASS : TAC_PLUS_AUTHEN_STATUS_FAIL;
	if (!res)
	    report(LOG_ERR, "dce auth error for %s", name);
	else
	    report(LOG_INFO, "dce auth ok for %s", name);
	return (res == 1);
    }
#endif

    /* Oops. No idea what kind of password this is. This should never
       happen as the parser should never create such passwords. */

    report(LOG_ERR, "%s: Error cannot identify password type %s for %s",
	   session.peer,
	   cfg_passwd && cfg_passwd[0] ? cfg_passwd : "<NULL>",
	   name ? name : "<unknown>");

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    return (0);
}

/* verify that this user/password is valid per /etc/passwd.
   Return 0 if invalid. */
static int
etc_passwd_file_verify(user, supplied_passwd, data, recurse)
char *user, *supplied_passwd;
struct authen_data *data;
int recurse;
{
    struct passwd *pw;
    char *exp_date;
    char *cfg_passwd;
#if defined(HAVE_SHADOW_H) || defined(SYSEXP)
    char buf[12];
#endif

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

#if defined(BSD_GETPWNAM) && (defined(TAC_PLUS_USERID) || defined(TAC_PLUS_USERNAME))
    seteuid(0);
#endif
    setpwent();
    pw = getpwnam(user);
    endpwent();
#if defined(BSD_GETPWNAM) && (defined(TAC_PLUS_USERID) || defined(TAC_PLUS_USERNAME))
    seteuid(TAC_PLUS_USERID);
#endif

    if (pw == NULL) {
	/* no entry exists */
	return (0);
    }

    if (*pw->pw_passwd == '\0' ||
	supplied_passwd == NULL ||
	*supplied_passwd == '\0') {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    cfg_passwd = pw->pw_passwd;
    exp_date = pw->pw_shell;

#ifdef HAVE_SHADOW_H
    if (STREQ(pw->pw_passwd, "x")) {
	struct spwd *spwd = getspnam(user);

	if (!spwd) {
	    if (debug & DEBUG_PASSWD_FLAG) {
		report(LOG_DEBUG, "No entry for %s in shadow file", user);
	    }
	    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	    return (0);
	}
	if (debug & DEBUG_PASSWD_FLAG) {
	    report(LOG_DEBUG, "Found entry for %s in shadow file", user);
	}
	cfg_passwd = spwd->sp_pwdp;

	/*
	 * Sigh. The Solaris shadow password file contains its own
	 * expiry date as the number of days after the epoch
	 * (January 1, 1970) when the password expires.
	 * Convert this to ascii so that the traditional tacacs
	 * password expiration routines work correctly.
	 */

	if (spwd->sp_expire > 0) {
#ifdef SHADEXP
            snprintf(buf, sizeof(buf), "%ld", spwd->sp_expire * 24 * 60 * 60);
#else
	    long secs = spwd->sp_expire * 24 * 60 * 60;
	    char *p = ctime(&secs);
	    bcopy(p+4, buf, 7);
	    bcopy(p+20, buf+7, 4);
	    buf[11] = '\0';
#endif
	    exp_date = buf;
	}
    }
#endif /* HAVE_SHADOW_H */

    /* try to verify the password */
    if (!des_verify(supplied_passwd, cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

#ifdef SYSEXP
    if(pw->pw_expire)
     {
      snprintf(buf, sizeof(buf), "%ld", pw->pw_expire);
      exp_date = buf;
     }
#endif

    /* password ok. Check expiry field */
    set_expiration_status(exp_date, data);

    if(!sys_adm_verify(user, pw, data, recurse))
      return 0;

    return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}

/* verify that this user/password is valid per a passwd(5) style
   database. Return 0 if invalid. */

static int
passwd_file_verify(user, supplied_passwd, data, filename, recurse)
char *user, *supplied_passwd;
struct authen_data *data;
char *filename;
int recurse;
{
    struct passwd *pw;
    char *exp_date;
    char *cfg_passwd;
#ifdef PWDDB
    char buf[12];
#endif

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    if (filename && STREQ(filename, "/etc/passwd")) {
	return(etc_passwd_file_verify(user, supplied_passwd, data, recurse));
    }

    /* an alternate filename */
    if (!(access(filename, R_OK) == 0)) {
	report(LOG_ERR, "%s %s: Cannot access %s for user %s -- %s",
	       session.peer, session.port, filename, user, strerror(errno));
	return (0);
    }

#ifdef PWDDB
    setpwfilet(filename);
    setpwentt();
    pw = getpwnamt(user);
    endpwentt();
#else
    pw = tac_passwd_lookup(user, filename);
#endif

    if (pw == NULL)
	/* no entry exists */
	return (0);

    if (*pw->pw_passwd == '\0' ||
	supplied_passwd == NULL ||
	*supplied_passwd == '\0') {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    cfg_passwd = pw->pw_passwd;
    exp_date = pw->pw_shell;

    /* try to verify the password */
    if (!des_verify(supplied_passwd, cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

#ifdef PWDDB
    if(pw->pw_expire)
     {
      snprintf(buf, sizeof(buf), "%ld", pw->pw_expire);
      exp_date = buf;
     }
#endif

    /* password ok. Check expiry field */
    set_expiration_status(exp_date, data);

    if(!sys_adm_verify(user, pw, data, recurse))
      return 0;

    return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}

/*
 * verify a provided password against a des encrypted one
 * return 1 if verified, 0 otherwise.
 */

int
des_verify(users_passwd, encrypted_passwd)
char *users_passwd, *encrypted_passwd;
{
    char *ep;

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "verify %s %s", users_passwd, encrypted_passwd);

    if (users_passwd == NULL ||
	*users_passwd == '\0' ||
	encrypted_passwd == NULL ||
	*encrypted_passwd == '\0') {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify returns 0");
	return (0);
    }

    ep = (char *) crypt(users_passwd, encrypted_passwd);

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "%s encrypts to %s", users_passwd, ep);

    if (strcmp(ep, encrypted_passwd) == 0) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password is correct");
	return (1);
    }

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "Password is incorrect");

    return (0);
}
